package com.hurlant.jsobject
{
	import flash.external.ExternalInterface;
	import flash.utils.Proxy;
	import flash.utils.flash_proxy;

	use namespace flash_proxy;

	/**
	 * 
	 * A proxy for the browser DOM, to let
	 * ecmascript code in the player feel just like it's 
	 * running with the browser environment.
	 * 
	 * @author henri
	 * 
	 */
	dynamic public class JSObject extends Proxy
	{
		internal var state:Object;
		internal var items:Array;
		private static var lock:Lock;
		 
		
		
		public function JSObject(obj:Object, l:Lock) {
			if (l==null) {
				throw "JSObject has a private constructor.";
			}
			state=obj;
		}
	
		static private var window:JSObject = null;	
		static private function init():void {
			lock = new Lock;
			ExternalInterface.addCallback("call", jsCall);
			ExternalInterface.call("HURLANT_jsobject_setup", ExternalInterface.objectID);
			window = new JSObject(ExternalInterface.call("HURLANT_jsobject_getReference","window"), lock);
		}
		static private var asfuncs:Object = {}
		static private function jsCall(n:String, that:*, a:*):* {
			var f:Function = asfuncs[n];
			if (f!=null) {
				for (var i:int=0;i<a.length;i++) {
					var ref:Object = a[i];
					a[i] = resolve(ref); 
				}
				var ret:* = f.apply(resolve(that), a);
				return wrap(ret);
			}
			return null;
		}
		static public function getWindow():JSObject {
			if (lock == null) {
				init();
			}
			return window;
		}
		static private var objMap:Object = {};
		static private function getFromRef(ref:Object):JSObject {
			if (objMap[ref.id]==null) {
				objMap[ref.id] = new JSObject(ref, lock);
			}
			return objMap[ref.id];
		}
		
		static private var idcount:int=0;
		static private function genId():String {
			return "aso"+(idcount++)
		}
		
		static private function getRef(name:String, parent:JSObject = null):* {
			var obj:Object = ExternalInterface.call("HURLANT_jsobject_getReference", name, parent?parent.state.id:null);
			return resolve(obj);
		}
		static private function setRef(name:String, parent:JSObject, value:*):void {
			ExternalInterface.call("HURLANT_jsobject_setRef", name, parent.state.id, wrap(value));
		}
		static private function wrap(value:*):* {
			if (value is JSObject) {
				return {type:"ref", id:value.state.id};
			} else if (value is Function) {
				var id:String = genId();
				asfuncs[id] = value;
				return {type:"func", id:id};
			} else {
				return {type:"value", val:value};
			}
		}
		static private function callRef(name:String, parent:JSObject, args:Array):* {
			// loop over arguments
			var hasJSObject:Boolean = false;
			for each (var a:* in args) {
				if (a is JSObject) {
					hasJSObject = true;
					break;
				}
			}
			if (hasJSObject) {
				for (var i:int=0;i<args.length;i++) {
					if (args[i] is JSObject) {
						args[i] = args[i].state.id;
					} else {
						args[i] = [args[i]];
					}
				}
				var v:* = ExternalInterface.call("HURLANT_jsobject_callRefById", name, parent.state.id, args);
			} else {
				v = ExternalInterface.call("HURLANT_jsobject_callRefByValue", name, parent.state.id, args); 
			}
			return resolve(v);
		} 
		static private function delRef(name:String, parent:JSObject):Boolean {
			return ExternalInterface.call("HURLANT_jsobject_delReference", name, parent.state.id);
		}
		static private function hasRef(name:String, parent:JSObject):Boolean {
			return ExternalInterface.call("HURLANT_jsobject_hasReference", name, parent.state.id);
		}
		static private function getIndexes(parent:JSObject):Array {
			return ExternalInterface.call("HURLANT_jsobject_getIndexes", parent.state.id);
		}
		static private function wrapJSFunc(obj:JSObject):* {
			var f:Object = function(...args):* {
				return ExternalInterface.call("HURLANT_jsobject_callRef", obj.state.id, args);
			};
			f.toString = function():String {
				return obj.toString();
			};
			return f;
		}
		static private function wrapASFunc(v:Function):* {
			var f:Object = function(a:*):* {
				for (var i:int=0;i<a.length;i++) {
					var ref:Object = a[i];
					a[i] = resolve(ref); 
				}
				var ret:* = v.apply(this,a);
				return wrap(ret);
			}
			return f;
		}
		static private function resolve(ref:Object):* {
			if (typeof ref.value != "undefined") return ref.value;
			var o:JSObject = getFromRef(ref);
			if (ref.callable == true) {
				return wrapJSFunc(o);
			}
			return o;
		}
		
		// the public API
		
		override flash_proxy function getProperty(name:*):* {
			return getRef(name, this);
		}
		
		override flash_proxy function setProperty(name:*, value:*):void {
			setRef(name, this, value);
		}
		
		override flash_proxy function callProperty(name:*, ...rest):* {
			return callRef(name, this, rest);
		}
		
		override flash_proxy function deleteProperty(name:*):Boolean {
			return delRef(name, this);
		}
		
		override flash_proxy function hasProperty(name:*):Boolean {
			return hasRef(name, this);
		}
		
		override flash_proxy function nextNameIndex(index:int):int {
			// initial call
			if (index==0) {
				items = getIndexes(this);
			}
			if (index < items.length) {
				return index+1;
			} else {
				return 0;
			}
		}
		override flash_proxy function nextName(index:int):String {
			return items[index-1];
		}
		override flash_proxy function nextValue(index:int):* {
			return getProperty(nextName(index));
		}
		
		
	}
}
class Lock {}